"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ExecutionEnvironment = void 0;
const child_process = require("child_process");
const fs = require("fs");
const path = require("path");
const vscode_uri_1 = require("vscode-uri");
const uuid_1 = require("uuid");
const imagePuller_1 = require("../utils/imagePuller");
const misc_1 = require("../utils/misc");
class ExecutionEnvironment {
    constructor(connection, context) {
        var _a;
        this.useProgressTracker = false;
        this.connection = connection;
        this.context = context;
        this.useProgressTracker =
            !!((_a = context.clientCapabilities.window) === null || _a === void 0 ? void 0 : _a.workDoneProgress);
    }
    initialize() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const settings = yield this.context.documentSettings.get(this.context.workspaceFolder.uri);
                if (!settings.executionEnvironment.enabled) {
                    return;
                }
                this._container_image = settings.executionEnvironment.image;
                this._container_engine = settings.executionEnvironment.containerEngine;
                if (this._container_engine === 'auto') {
                    for (const ce of ['podman', 'docker']) {
                        try {
                            child_process.execSync(`which ${ce}`, {
                                encoding: 'utf-8',
                            });
                        }
                        catch (error) {
                            this.connection.console.info(`Container engine '${ce}' not found`);
                            continue;
                        }
                        this._container_engine = ce;
                        this.connection.console.log(`Container engine set to: '${ce}'`);
                        break;
                    }
                }
                else {
                    try {
                        child_process.execSync(`which ${this._container_engine}`, {
                            encoding: 'utf-8',
                        });
                    }
                    catch (error) {
                        this.connection.window.showErrorMessage(`Container engine '${this._container_engine}' not found. Failed with error '${error}'`);
                        return;
                    }
                }
                if (!['podman', 'docker'].includes(this._container_engine)) {
                    this.connection.window.showInformationMessage('No valid container engine found.');
                    return;
                }
                const imagePuller = new imagePuller_1.ImagePuller(this.connection, this.context, this._container_engine, this._container_image, settings.executionEnvironment.pullPolicy);
                const setupDone = yield imagePuller.setupImage();
                if (!setupDone) {
                    this.connection.window.showErrorMessage(`Execution environment image '${this._container_image}' setup failed.
           For more details check output console logs for ansible-language-server`);
                    return;
                }
                this.fetchPluginDocs();
            }
            catch (error) {
                if (error instanceof Error) {
                    this.connection.window.showErrorMessage(error.message);
                }
                else {
                    this.connection.console.error(`Exception in ExecutionEnvironment service: ${JSON.stringify(error)}`);
                }
            }
        });
    }
    fetchPluginDocs() {
        return __awaiter(this, void 0, void 0, function* () {
            const ansibleConfig = yield this.context.ansibleConfig;
            const containerName = `${this._container_image.replace(/[^a-z0-9]/gi, '_')}`;
            let progressTracker;
            try {
                const containerImageIdCommand = `${this._container_engine} images ${this._container_image} --format="{{.ID}}" | head -n 1`;
                this.connection.console.log(containerImageIdCommand);
                this._container_image_id = child_process
                    .execSync(containerImageIdCommand, {
                    encoding: 'utf-8',
                })
                    .trim();
                const hostCacheBasePath = path.resolve(`${process.env.HOME}/.cache/ansible-language-server/${containerName}/${this._container_image_id}`);
                const isContainerRunning = this.runContainer(containerName);
                if (!isContainerRunning) {
                    return;
                }
                if (fs.existsSync(hostCacheBasePath)) {
                    ansibleConfig.collections_paths = this.updateCachePaths(ansibleConfig.collections_paths, hostCacheBasePath);
                    ansibleConfig.module_locations = this.updateCachePaths(ansibleConfig.module_locations, hostCacheBasePath);
                }
                else {
                    if (this.useProgressTracker) {
                        progressTracker =
                            yield this.connection.window.createWorkDoneProgress();
                    }
                    if (progressTracker) {
                        progressTracker.begin('execution-environment', undefined, `Copy plugin docs from '${this._container_image} to host cache path`, true);
                    }
                    ansibleConfig.collections_paths = yield this.copyPluginDocFiles(hostCacheBasePath, containerName, ansibleConfig.collections_paths, '**/ansible_collections');
                    const builtin_plugin_locations = [];
                    ansibleConfig.module_locations.forEach((modulePath) => {
                        const pluginsPathParts = modulePath.split(path.sep).slice(0, -1);
                        if (pluginsPathParts.includes('site-packages')) {
                            // ansible-config returns default builtin configured module path
                            // as ``<python-path>/site-packages/ansible/modules`` to copy other plugins
                            // to local cache strip the ``modules`` part from the path and append
                            // ``plugins`` folder.
                            pluginsPathParts.push('plugins');
                        }
                        builtin_plugin_locations.push(pluginsPathParts.join(path.sep));
                    });
                    // Copy builtin plugins
                    yield this.copyPluginDocFiles(hostCacheBasePath, containerName, builtin_plugin_locations, '*');
                    // Copy builtin modules
                    ansibleConfig.module_locations = yield this.copyPluginDocFiles(hostCacheBasePath, containerName, ansibleConfig.module_locations, '**/modules');
                }
            }
            catch (error) {
                this.connection.window.showErrorMessage(`Exception in ExecutionEnvironment service while fetching docs: ${JSON.stringify(error)}`);
            }
            finally {
                if (progressTracker) {
                    progressTracker.done();
                }
                this.cleanUpContainer(containerName);
            }
        });
    }
    wrapContainerArgs(command, mountPaths) {
        const workspaceFolderPath = vscode_uri_1.URI.parse(this.context.workspaceFolder.uri).path;
        const containerCommand = [this._container_engine];
        containerCommand.push(...['run', '--rm']);
        containerCommand.push(...['--workdir', workspaceFolderPath]);
        containerCommand.push(...['-v', `${workspaceFolderPath}:${workspaceFolderPath}`]);
        for (const mountPath of mountPaths || []) {
            const volumeMountPath = `${mountPath}:${mountPath}`;
            if (containerCommand.includes(volumeMountPath)) {
                continue;
            }
            containerCommand.push('-v', volumeMountPath);
        }
        if (this._container_engine === 'podman') {
            // container namespace stuff
            containerCommand.push('--group-add=root');
            containerCommand.push('--ipc=host');
            // docker does not support this option
            containerCommand.push('--quiet');
        }
        else {
            containerCommand.push(`--user=${process.getuid()}`);
        }
        containerCommand.push(`--name ansible_language_server_${(0, uuid_1.v4)()}`);
        containerCommand.push(this._container_image);
        containerCommand.push(command);
        const generatedCommand = containerCommand.join(' ');
        this.connection.console.log(`container engine invocation: ${generatedCommand}`);
        return generatedCommand;
    }
    cleanUpContainer(containerName) {
        [
            `${this._container_engine} stop ${containerName}`,
            `${this._container_engine} rm ${containerName}`,
        ].forEach((command) => {
            try {
                child_process.execSync(command, {
                    cwd: vscode_uri_1.URI.parse(this.context.workspaceFolder.uri).path,
                });
            }
            catch (error) {
                // container already stopped and/or removed
            }
        });
    }
    isPluginInPath(containerName, searchPath, pluginFolderPath) {
        const command = `${this._container_engine} exec ${containerName} find ${searchPath} -path '${pluginFolderPath}'`;
        try {
            this.connection.console.info(`Executing command ${command}`);
            const result = child_process
                .execSync(command, {
                encoding: 'utf-8',
            })
                .trim();
            return result !== '';
        }
        catch (error) {
            this.connection.console.error(error);
            return false;
        }
    }
    runContainer(containerName) {
        // ensure container is not running
        this.cleanUpContainer(containerName);
        try {
            const command = `${this._container_engine} run --rm -it -d --name ${containerName} ${this._container_image} bash`;
            this.connection.console.log(`run container with command '${command}'`);
            child_process.execSync(command, {
                encoding: 'utf-8',
            });
        }
        catch (error) {
            this.connection.window.showErrorMessage(`Failed to initialize execution environment '${this._container_image}': ${error}`);
            return false;
        }
        return true;
    }
    copyPluginDocFiles(hostPluginDocCacheBasePath, containerName, containerPluginPaths, searchKind) {
        return __awaiter(this, void 0, void 0, function* () {
            const updatedHostDocPath = [];
            if (fs.existsSync(hostPluginDocCacheBasePath)) {
                containerPluginPaths.forEach((srcPath) => {
                    updatedHostDocPath.push(path.join(hostPluginDocCacheBasePath, srcPath));
                });
            }
            else {
                containerPluginPaths.forEach((srcPath) => {
                    if (srcPath === '' ||
                        !this.isPluginInPath(containerName, srcPath, searchKind)) {
                        return;
                    }
                    const destPath = path.join(hostPluginDocCacheBasePath, srcPath);
                    const destPathFolder = destPath
                        .split(path.sep)
                        .slice(0, -1)
                        .join(path.sep);
                    fs.mkdirSync(destPath, { recursive: true });
                    const copyCommand = `docker cp ${containerName}:${srcPath} ${destPathFolder}`;
                    this.connection.console.log(`Copying plugins from container to local cache path ${copyCommand}`);
                    (0, misc_1.asyncExec)(copyCommand, {
                        encoding: 'utf-8',
                    });
                    updatedHostDocPath.push(destPath);
                });
            }
            return updatedHostDocPath;
        });
    }
    updateCachePaths(pluginPaths, cacheBasePath) {
        const localCachePaths = [];
        pluginPaths.forEach((srcPath) => {
            const destPath = path.join(cacheBasePath, srcPath);
            if (fs.existsSync(destPath)) {
                localCachePaths.push(destPath);
            }
        });
        return localCachePaths;
    }
}
exports.ExecutionEnvironment = ExecutionEnvironment;
//# sourceMappingURL=executionEnvironment.js.map